import ij.*;
import ij.plugin.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.awt.image.*;


public class CDgetPrimaryChannel implements PlugIn {

/**
 *  CDgetPrimaryChannel.java
 *  @version 1.1
 *  @author e.a.papp@medisin.uio.no 
 *
 *  This is a modified version of the original Colour Deconvolution plugin
 *  by G. Landini based on the method of A.C. Ruifrok. 
 *  @see http://imagejdocu.tudor.lu/doku.php?id=plugin:color:colour_deconvolution:start
 *
 *  The original plugin was modified for use with batch processing of series of images:
 *   - CD vectors are passed to the plugin as a parameter string
 *   - Only the primary image channel is calculated
 *   - No dialogs or windows are displayed
 * 
 *  For full functionality, e.g. acquisition of new vectors, please follow the above link
 *  and install the original plugin.
 * 
 *  The modified plugin is designed to be used together with the NeSys Labeling Detection Macro
 *  (NeSys_Labeling_Detection_Macro.txt in the plugins/Macros directory) and the 
 *  colourdeconvolution.txt parameter file for storing vectors.
 *   
 *  This plugin is part of a workflow for automated detection of axonal labeling 
 *  in series of microscopic images. See the ImageJ DokuWiki for details:
 *  http://imagejdocu.tudor.lu/doku.php?id=plugin:analysis:autodetection_of_neuronal_labeling_in_histological_image_series:start
 */

// G.Landini at bham ac uk
// 30/Mar/2004 released
// 03/Apr/2004 resolved ROI exiting
// 07/Apr/2004 added Methyl Green DAB vectors
// 08/Jul/2004 shortened the code
// 01/Aug/2005 added fast red/blue/DAB vectors
// 02/Nov/2005 changed code to work with image stacks (DLC - dchao at fhcrc org)
// 02/Nov/2005 changed field names so user-defined colours can be set within 
//             macros (DLC - dchao at fhcrc org)
// 04/Feb/2007 1.3 disable popup menu when right clicking
//
// This plugin implements stain separation using the colour deconvolution
// method described in:
//
//     Ruifrok AC, Johnston DA. Quantification of histochemical
//     staining by color deconvolution. Analytical & Quantitative
//     Cytology & Histology 2001; 23: 291-299.
//
// The code is based on "Color separation-30", a macro for NIH Image kindly provided
// by A.C. Ruifrok. Thanks Arnout!
//
// The plugin assumes images generated by color subtraction (i.e. light-absorbing dyes
// such as those used in bright field histology or ink on printed paper) but the dyes
// should not be neutral grey.
//
// I strongly suggest to read the paper reference above to understand how to determine
// new vectors and how the whole procedure works.
//
// The plugin works correctly when the background is neutral (white to light grey), 
// so background subtraction and colour correction must be applied to the images before 
// processing.
//
// The plugin provides a number of "built in" stain vectors some of which were determined
// experimentally in our lab (marked GL), but you may have to determine your own vectors to
// provide a more accurate stain separation, depending on the stains and methods you use.
// Ideally, vector determination should be done on slides stained with only one colour
// at a time (using the "From ROI" interactive option).
//
// The plugin takes an RGB image and returns three 8-bit images. If the specimen is
// stained with a 2 colour scheme (such as H & E) the 3rd image represents the
// complimentary of the first two colours (i.e. green).
//
// Please be *very* careful about how to interpret the results of colour deconvolution
// when analysing histological images.
// Most staining methods are not stochiometric and so optical density of the chromogen
// may not correlate well with the quantity of the reactants.
// This means that optical density of the colour may not be a good indicator of
// the amount of material stained.
//
// Read the paper!
//

	public static void startCD(String arg){

		CDgetPrimaryChannel cpc = new CDgetPrimaryChannel();
	        cpc.run(arg);
	} 

	public void run(String arg) {

		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp==null){
			IJ.error("No image!");
			return;
		}
		int width = imp.getWidth();
		int height = imp.getHeight();
                String title = imp.getTitle();

		double leng, A, V, C, log255=Math.log(255.0);
		int i,j;
		double [] MODx = new double[3];
		double [] MODy = new double[3];
		double [] MODz = new double[3];
		double [] cosx = new double[3];
		double [] cosy = new double[3];
		double [] cosz  = new double[3];
		double [] len = new double[3];
		double [] q = new double[9];
		byte [] rLUT = new byte[256];
		byte [] gLUT = new byte[256];
		byte [] bLUT = new byte[256];


		String [] CDvectors = arg.split(",");

		MODx[0] = Double.parseDouble(CDvectors[1]);
		MODy[0] = Double.parseDouble(CDvectors[2]);
		MODz[0] = Double.parseDouble(CDvectors[3]);

		MODx[1] = Double.parseDouble(CDvectors[4]);
		MODy[1] = Double.parseDouble(CDvectors[5]);
		MODz[1] = Double.parseDouble(CDvectors[6]);	

		MODx[2] = Double.parseDouble(CDvectors[7]);
		MODy[2] = Double.parseDouble(CDvectors[8]);
		MODz[2] = Double.parseDouble(CDvectors[9]);

		// start
		for (i=0; i<3; i++){
			//normalise vector length
			cosx[i]=cosy[i]=cosz[i]=0.0;
			len[i]=Math.sqrt(MODx[i]*MODx[i] + MODy[i]*MODy[i] + MODz[i]*MODz[i]);
			if (len[i] != 0.0){
				cosx[i]= MODx[i]/len[i];
				cosy[i]= MODy[i]/len[i];
				cosz[i]= MODz[i]/len[i];
			}
		}


		// translation matrix
		if (cosx[1]==0.0){ //2nd colour is unspecified
			if (cosy[1]==0.0){
				if (cosz[1]==0.0){
					cosx[1]=cosz[0];
					cosy[1]=cosx[0];
					cosz[1]=cosy[0];
				}
			}
		}

		if (cosx[2]==0.0){ // 3rd colour is unspecified
			if (cosy[2]==0.0){
				if (cosz[2]==0.0){
					if ((cosx[0]*cosx[0] + cosx[1]*cosx[1])> 1){
						cosx[2]=0.0;
					}
					else {
						cosx[2]=Math.sqrt(1.0-(cosx[0]*cosx[0])-(cosx[1]*cosx[1]));
					}

					if ((cosy[0]*cosy[0] + cosy[1]*cosy[1])> 1){
						cosy[2]=0.0;
					}
					else {
						cosy[2]=Math.sqrt(1.0-(cosy[0]*cosy[0])-(cosy[1]*cosy[1]));
					}

					if ((cosz[0]*cosz[0] + cosz[1]*cosz[1])> 1){
						cosz[2]=0.0;
					}
					else {
						cosz[2]=Math.sqrt(1.0-(cosz[0]*cosz[0])-(cosz[1]*cosz[1]));
					}
				}
			}
		}

		leng=Math.sqrt(cosx[2]*cosx[2] + cosy[2]*cosy[2] + cosz[2]*cosz[2]);

		cosx[2]= cosx[2]/leng;
		cosy[2]= cosy[2]/leng;
		cosz[2]= cosz[2]/leng;
				
		//matrix inversion
		A = cosy[1] - cosx[1] * cosy[0] / cosx[0];
		V = cosz[1] - cosx[1] * cosz[0] / cosx[0];
		C = cosz[2] - cosy[2] * V/A + cosx[2] * (V/A * cosy[0] / cosx[0] - cosz[0] / cosx[0]);
		q[2] = (-cosx[2] / cosx[0] - cosx[2] / A * cosx[1] / cosx[0] * cosy[0] / cosx[0] + cosy[2] / A * cosx[1] / cosx[0]) / C;
		q[1] = -q[2] * V / A - cosx[1] / (cosx[0] * A);
		q[0] = 1.0 / cosx[0] - q[1] * cosy[0] / cosx[0] - q[2] * cosz[0] / cosx[0];
		q[5] = (-cosy[2] / A + cosx[2] / A * cosy[0] / cosx[0]) / C;
		q[4] = -q[5] * V / A + 1.0 / A;
		q[3] = -q[4] * cosy[0] / cosx[0] - q[5] * cosz[0] / cosx[0];
		q[8] = 1.0 / C;
		q[7] = -q[8] * V / A;
		q[6] = -q[7] * cosy[0] / cosx[0] - q[8] * cosz[0] / cosx[0];


		// initialize processor for output image
		ByteProcessor outputip = new ByteProcessor(width, height);

		for (j=0; j<256; j++) { //LUT[1]

			rLUT[255-j]=(byte)(255.0 - (double)j * cosx[0]);
			gLUT[255-j]=(byte)(255.0 - (double)j * cosy[0]);
			bLUT[255-j]=(byte)(255.0 - (double)j * cosz[0]);
		}
		IndexColorModel cm = new IndexColorModel(8, 256, rLUT, gLUT, bLUT);
		outputip.setColorModel(cm);

		// translate
                int imagesize = width * height;
                int[] pixels = (int[])imp.getProcessor().getPixels();
                byte[] newpixels = new byte[imagesize];

                for (j=0;j<imagesize;j++){

                	// log transform the RGB data
	                int R = (pixels[j] & 0xff0000)>>16;
    		        int G = (pixels[j] & 0x00ff00)>>8 ;
                        int B = (pixels[j] & 0x0000ff);
                        double Rlog = -((255.0*Math.log(((double)R+1)/255.0))/log255);
                        double Glog = -((255.0*Math.log(((double)G+1)/255.0))/log255);
                        double Blog = -((255.0*Math.log(((double)B+1)/255.0))/log255);

                        // rescale to match original paper values
                        double Rscaled = Rlog * q[0*3];
                        double Gscaled = Glog * q[0*3+1];
                        double Bscaled = Blog * q[0*3+2];
                        double output = Math.exp(-((Rscaled + Gscaled + Bscaled) - 255.0) * log255 / 255.0);

                        if(output>255) output=255;
                        newpixels[j]=(byte)(0xff&(int)(Math.floor(output+.5)));

                        // add new values to output image
   		        outputip.setPixels(newpixels);
                }

		String newTitle = title;

		if (title.contains(".")) newTitle = title.substring(0, title.lastIndexOf('.'));
               
                new ImagePlus(newTitle + "_binary", outputip).show();
	}
}

